package com.ibm.rules.support;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import com.ibm.rules.decisioncenter.BRMServerClientLegacy;

import ilog.rules.teamserver.brm.IlrBOMPathEntry;
import ilog.rules.teamserver.brm.IlrBaseline;
import ilog.rules.teamserver.brm.IlrBrmPackage;
import ilog.rules.teamserver.brm.IlrDependency;
import ilog.rules.teamserver.brm.IlrProjectBOMEntry;
import ilog.rules.teamserver.brm.IlrProjectInfo;
import ilog.rules.teamserver.brm.IlrRuleProject;
import ilog.rules.teamserver.model.IlrCommitableObject;
import ilog.rules.teamserver.model.IlrObjectNotFoundException;
import ilog.rules.teamserver.model.IlrSessionEx;
import ilog.rules.teamserver.model.IlrSessionHelper;
import ilog.rules.teamserver.model.permissions.IlrPermissionException;
import ilog.rules.teamserver.validation.IlrValidationPackage;

public class BomPathRepair extends BRMServerClientLegacy {
	public String project;
	public IlrValidationPackage valPackage;
	public IlrBrmPackage brm;

	protected void init() throws Exception {
		this.valPackage = getSession().getValidationPackage();
		this.brm = getSession().getBrmPackage();
	}

	public void inspectRepository(String dsName, boolean fix) throws IlrObjectNotFoundException {
		try {
			init();
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
		if ((dsName != null) && (!dsName.isEmpty())) {
			IlrRuleProject dsProject = IlrSessionHelper.getProjectNamed(getSession(), dsName);
			if (dsProject == null) {
				System.err.println("ERROR: no project '" + dsName + "' found !");
				return;
			}
			checkProject(dsProject, fix);
		} else {
			List<IlrRuleProject> allProjects = IlrSessionHelper.getProjects(getSession());
			for (IlrRuleProject ilrRuleProject : allProjects) {
				checkProject(ilrRuleProject, fix);
			}
		}
	}

	private void checkProject(IlrRuleProject dsProject, boolean fix) {
		try {
			List<IlrBaseline> baselines;
			baselines = IlrSessionHelper.getBaselines(getSession(), dsProject);
			System.out.println("Checking " + dsProject.getName() + "(baselines " + baselines.size() + ")");

			for (IlrBaseline ilrBaseline : baselines) {

				getSession().setWorkingBaseline(ilrBaseline);

				IlrCommitableObject co = null;
				int order = 0;
				IlrProjectInfo projectinfo = ilrBaseline.getProjectInfo();
				List<IlrDependency> dependencies = projectinfo.getDependencies();
				ArrayList<String> dependentProjects = new ArrayList<>();
				ArrayList<String> bomProjects = new ArrayList<>();
				for (IlrDependency ilrDependency : dependencies) {
					dependentProjects.add(ilrDependency.getProjectName());
				}
				List<IlrBOMPathEntry> bompaths = projectinfo.getBomPathEntries();
				for (IlrBOMPathEntry ilrBOMPathEntry : bompaths) {
					if (ilrBOMPathEntry instanceof IlrProjectBOMEntry) {
						bomProjects.add(((IlrProjectBOMEntry) ilrBOMPathEntry).getProjectName());
					}
					order++;
				}
				if (fix) {
					co = new IlrCommitableObject(projectinfo);
					co.setRootDetails(projectinfo);
				}
				// search if each BOM path entry that is a project match a project dependency
				for (IlrBOMPathEntry bomProjectEntry : bompaths) {
					if (!(bomProjectEntry instanceof IlrProjectBOMEntry))
						continue;
					String bomProject = ((IlrProjectBOMEntry) bomProjectEntry).getProjectName();
					if (!dependentProjects.contains(bomProject)) {
						String branchName = ilrBaseline.getName();
						if (fix) {
							if (existsBaseline(bomProject, branchName)) {
								System.out.println(MessageFormat.format(
										"Project:{0} branch:{1}: adding dependency to BOM project {2}",
										dsProject.getName(), branchName, bomProject));
								IlrDependency dependency = (IlrDependency) ((IlrSessionEx) getSession())
										.createElementDetails(brm.getDependency());
								dependency.setProjectName(bomProject);
								dependency.setBaselineName(branchName);
								co.addModifiedElement(brm.getProjectInfo_Dependencies(), dependency);
							} else { // remove the bom path entry
								System.out.println(MessageFormat.format(
										"Project:{0} branch:{1}: removing BOM project {2} from BOM path, branch does not exist",
										dsProject.getName(), branchName, bomProject));
								co.addDeletedElement(brm.getProjectInfo_BomPathEntries(), bomProjectEntry);
							}
						} else {
							System.out.println(MessageFormat.format(
									"Project:{0} branch:{1}: reference BOM project {2} and does not have a dependency on it",
									dsProject.getName(), branchName, bomProject));

						}
					}
				}

				// search if all dependent projects are listed in the BOM path
				for (String dependentProject : dependentProjects) {
					if (!bomProjects.contains(dependentProject)) {
						if (fix) {
							System.out.println(MessageFormat.format(
									"Project:{0} branch:{1}: inserting BOM path entry for project dependency {2}",
									dsProject.getName(), ilrBaseline.getName(), dependentProject));
							IlrBOMPathEntry entry = (IlrProjectBOMEntry) ((IlrSessionEx) getSession())
									.createElementDetails(brm.getProjectBOMEntry());
							entry.setRawValue(brm.getProjectBOMEntry_ProjectName(), dependentProject);
							entry.setRawValue(brm.getBOMPathEntry_Url(), "platform:/" + dependentProject);
							entry.setRawValue(brm.getBOMPathEntry_Order(), order++);
							co.addModifiedElement(brm.getProjectInfo_BomPathEntries(), entry);

						} else {
							System.out.println(MessageFormat.format(
									"Project:{0} branch:{1}: has a dependency on project {2} and does not have a BOM entry for it",
									dsProject.getName(), ilrBaseline.getName(), dependentProject));

						}
					}
				}
				// save the changes in the project info
				if (fix) {
					try {
						getSession().commit(co);
					} catch (Throwable e) {
						System.out.println("unable to save the project info updates");
						e.printStackTrace();
					}
				}
			}
		} catch (

		IlrObjectNotFoundException e) {
			e.printStackTrace();
		} catch (IlrPermissionException e1) {
			e1.printStackTrace();
		}
	}

	private boolean existsBaseline(String bomProject, String branchName) {
		try {
			IlrRuleProject project = IlrSessionHelper.getProjectNamed(getSession(), bomProject);
			if (project == null)
				return false;
			IlrBaseline baseline = IlrSessionHelper.getBaselineNamed(getSession(), project, branchName);
			return baseline != null;
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
			return false;
		}
	}

	public static void main(final String[] args) {
		try {
			(new BomPathRepair()).doExecute(args);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	protected void execute(String[] otherArgs) {
		String prjName = null;
		boolean fixMode = false;
		int i = 0;
		int valueIndex = -1; 
		for (String param : otherArgs) {
			if (valueIndex ==i)
			{
				i++;
				continue;
			}
			switch (param) {
			case PRJ_ARGNAME:
				valueIndex = i+1;
				if (valueIndex >= otherArgs.length) {
					System.out.println("ERROR: wrong number of arguments; the name of the project is missing");
					usage();
					return;
				}
				prjName = otherArgs[valueIndex];
				break;
			case FIX_MODE:
				valueIndex = i+1;
				if (valueIndex < otherArgs.length) {
					fixMode = otherArgs[valueIndex].equalsIgnoreCase("true");
				}
				break;
			case TOKEN:
				valueIndex = i+1;
				break;
			case ZENAPIKEY:
				valueIndex = i+1;
				break;
			default:
				System.out.println("ERROR: unexpected argument " + param);
				usage();
				return;
			}
			i++;
		}
		
		try {
				inspectRepository(prjName, fixMode);

			} catch (Exception e){
				e.printStackTrace();
			}
	}
	
	private final static String PRJ_ARGNAME = "-project";
	private final static String FIX_MODE = "-fix";
	private final static String TOKEN = "-token";
	private final static String ZENAPIKEY = "-zenApiKey";

	@Override
	protected String usageArgs() {
		return PRJ_ARGNAME + " <dsPrjName> " + FIX_MODE + " true|false"
				+ "\tcheck the bom path dependencies in all project  or a given project display the problems and repair if fix mode is true";
	}
}
