package com.ibm.rules.support;

import ilog.rules.teamserver.brm.IlrActivity;
import ilog.rules.teamserver.brm.IlrBaseline;
import ilog.rules.teamserver.brm.IlrBranch;
import ilog.rules.teamserver.brm.IlrBrmPackage;
import ilog.rules.teamserver.brm.IlrChangeAct;
import ilog.rules.teamserver.brm.IlrDependency;
import ilog.rules.teamserver.brm.IlrModelElement;
import ilog.rules.teamserver.brm.IlrProjectInfo;
import ilog.rules.teamserver.brm.IlrRelease;
import ilog.rules.teamserver.brm.IlrRuleProject;
import ilog.rules.teamserver.brm.impl.IlrDependencyImpl;
import ilog.rules.teamserver.brm.impl.IlrModelElementImpl;
import ilog.rules.teamserver.client.IlrRemoteSessionFactory;
import ilog.rules.teamserver.model.IlrApplicationException;
import ilog.rules.teamserver.model.IlrCommitableObject;
import ilog.rules.teamserver.model.IlrConnectException;
import ilog.rules.teamserver.model.IlrCyclicBaselineDependencyException;
import ilog.rules.teamserver.model.IlrElementHandle;
import ilog.rules.teamserver.model.IlrModelInfo;
import ilog.rules.teamserver.model.IlrObjectNotFoundException;
import ilog.rules.teamserver.model.IlrSession;
import ilog.rules.teamserver.model.IlrSessionEx;
import ilog.rules.teamserver.model.IlrSessionFactory;
import ilog.rules.teamserver.model.IlrSessionHelper;
import ilog.rules.teamserver.model.IlrSessionHelperEx;
import ilog.rules.teamserver.model.finders.DataFinder;
import ilog.rules.teamserver.model.governance.ReleaseController;
import ilog.rules.teamserver.model.impl.IlrDecisionServiceHelper;
import ilog.rules.teamserver.model.permissions.IlrPermissionException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;


public class CheckCircularDependencies {
	boolean dgfOverride = true;
	boolean debug = false;
	protected IlrSession session;
	private static String version = "V1.0";
	static Map<String, Map<String, List<String>> > projectPaths = new HashMap<>();

	public CheckCircularDependencies(IlrSession session) {
		this.session = session;
		System.out.println("Check Circular Dependencies " + version);
	}
	public List<String> getParentProjects(IlrSession session,IlrProjectInfo projectinfo, String depProject) {
		List<String> result = new ArrayList<String>();
		if (depProject.isEmpty())
			return result;
		try {
			
			List<IlrDependencyImpl> releaseprojects = projectinfo.getDependencies();

			for (IlrDependencyImpl project : releaseprojects) {
				if (depProject.equals(project.getProjectName())) {
					result.add(projectinfo.getProject().getName());
				}
			}

			for (IlrDependencyImpl project2 : releaseprojects) {

				String baselinename = project2.getBaselineName();
				String projectname = project2.getProjectName();
				IlrRuleProject projectDep = IlrSessionHelper.getProjectNamed(session, projectname);
				IlrBaseline baseline = IlrSessionHelper.getBaselineNamed(session, projectDep, baselinename);
				IlrProjectInfo info = baseline.getProjectInfo();

				result.addAll(getParentProjects(session, info, depProject));
			}
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
		}
		result = new ArrayList<String>(new HashSet<String>(result));
		return result;
	}

	public void inspectRepository(String dsName, String branch,Boolean fix)throws IlrObjectNotFoundException {
		if ((dsName != null) && (!dsName.isEmpty())) { //check given decision service
			IlrRuleProject dsProject = IlrSessionHelper.getProjectNamed(this.session, dsName);
				checkProject(dsProject,branch,fix);
		} else {//check all decision services
			List<IlrRuleProject> allProjects = IlrSessionHelper.getProjects(this.session);
			for (IlrRuleProject ilrRuleProject : allProjects) {
				System.out.println("Checking "+ilrRuleProject.getName());
				checkProject(ilrRuleProject,branch,fix);
			}
		}
	}

	public void checkProject(IlrRuleProject ruleproject, String branch, Boolean fix) { //check to run for releases or branches
		try {
			if (ruleproject.isDecisionService()) {
				if (IlrDecisionServiceHelper.isDGFProject(session, ruleproject)&&!dgfOverride) {
					checkProjectDependenciesInReleases(ruleproject,branch,fix);
				} else {
					checkProjectDependenciesInBranches(ruleproject,branch,fix);
				}
			}
			//checkReleaseParent(ruleproject);

		} catch (Exception e) {
			System.out.println("Project: " + (ruleproject != null ? (ruleproject.getName() + "Unable to process this project"): "Not found"));
			e.printStackTrace();
		}
	}
	
	public void printProjectPaths(String n, Boolean fix) throws IlrObjectNotFoundException, IlrApplicationException{
	
	for (String p : projectPaths.keySet()) {
		String o="  ";
		if(projectPaths.get(p).keySet().size()>1){
			System.out.println("Found multiple versions of dependency "+p+" in "+n);
			String mostOccuringBranch=null;
			int mostOccurances=0;
			for (String b : projectPaths.get(p).keySet()){
				int occurances=projectPaths.get(p).get(b).size();
				if (occurances>mostOccurances){
					mostOccurances=occurances;
					mostOccuringBranch=b;
				}
				for (String path : projectPaths.get(p).get(b)){
					System.out.println((new String(new char[2]).replace("\0", o))+"Branch=\""+b+ "\" path="+path);
				}
			}
		}
	}
}

public void renameBranch( IlrBaseline srcBranch, String newBranchName, List<String>projectsProcessed) throws IlrApplicationException{
	IlrProjectInfo pi = srcBranch.getProjectInfo();
	IlrBrmPackage brmPackage = session.getBrmPackage();
	IlrCommitableObject copi = new IlrCommitableObject(pi);
	String srcBranchName = srcBranch.getName();
	projectsProcessed.add(srcBranch.getProject().getName());
	List<IlrDependency> deps = pi.getDependencies();
	for (IlrDependency ilrDependency : deps) {
		IlrRuleProject depProject = IlrSessionHelper.getProjectNamed(session, ilrDependency.getProjectName());
		// skip if project not found
		if (depProject == null ){
			
			System.err.println("Unable to find dependent project "+ ilrDependency.getProjectName());
			continue;
		}
		IlrBaseline  destBranch = IlrSessionHelper.getBaselineNamed(session, depProject, newBranchName);
		if (destBranch == null){
			// if branch does not exists create it b renaming the srcBranch
			IlrBaseline  depBranch = IlrSessionHelper.getBaselineNamed(session, depProject, srcBranchName);
			if (depBranch == null){
				System.err.println("Unable to find "+srcBranchName+" in project "+ilrDependency.getProjectName());
				continue;
			}
			renameBranch(depBranch, newBranchName, projectsProcessed);
		}
		//should delete the srcBranch if the new branch name already exists.
		// update the name of the branch name in the dependencies of the current project
		ilrDependency.setBaselineName(newBranchName);
		copi.addModifiedElement(brmPackage.getProjectInfo_Dependencies(), ilrDependency);
	}
	        session.commit(copi);
	        // rename the temporary branch into release
	        srcBranch.setName(newBranchName);
	        IlrCommitableObject co = new IlrCommitableObject(srcBranch.getProject());
	        co.addModifiedElement(brmPackage.getRuleProject_Baselines(), srcBranch);
	        session.commit(co);
	}

	private void checkProjectDependenciesInBranches(IlrRuleProject ruleproject, String branch,boolean fix) {
		try {
			
			projectPaths = new HashMap<>();
			
			if ((branch != null) && (!branch.isEmpty())){ //if branch is defined, only search that branch
				IlrBaseline Branch = IlrSessionHelper.getBaselineNamed(session, ruleproject, branch) ;
				List<String> DependentProjects = getDependenciesFromBaseline(Branch,ruleproject.getName(),fix);
				//printProjectPaths(ruleproject.getName()+':'+Branch.getName(),fix);
			}
			
			else{ //Run for all branches
				IlrBaseline mainBranch = IlrSessionHelper.getCurrentBaseline(session, ruleproject) ;
					List<String> mainDependentProjects = getDependenciesFromBaseline(mainBranch,ruleproject.getName(),fix);
					List<IlrBaseline> branches = IlrSessionHelper.getBaselines(session, ruleproject);
			
					//printProjectPaths(ruleproject.getName()+':'+mainBranch.getName(),fix);
					for (IlrBaseline baseline: branches) {
						projectPaths = new HashMap<>();
						if (!baseline.equals( mainBranch) && !IlrDecisionServiceHelper.isDGFBranch(baseline) && baseline.isBranch() )
						{
							List<String> baselineDependentProjects = getDependenciesFromBaseline(baseline,ruleproject.getName(),fix);
							//printProjectPaths(ruleproject.getName()+':'+baseline.getName(),fix);
						}
					}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}

	public void checkReleaseParent(IlrRuleProject dsProject) {
		try {
			List<IlrBaseline> baselines = IlrSessionHelper.getBaselines(session, dsProject);
			for (IlrBaseline baseline : baselines) {
				// searching for release
				if (baseline.isDecisionServiceBranch()&& session.getBrmPackage().getRelease().isSuperTypeOf(baseline.eClass())) {
					IlrRelease release = (IlrRelease) baseline;
					if (release.isInitialRelease()&& !release.getName().contains("odm_initialRelease_")) {
						if (release.getParent() != null&& release.getParent().getParent() == null) {
							System.err.println("* release "+ release.getName()+ " does not have Initial Release for parent");
							fixReleaseParent(dsProject, release);
						}
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void fixReleaseParent(IlrRuleProject project, IlrRelease release) {
		try {
			IlrRuleProject service = findDecisionService(project, release);
			if (service == null) {
				System.err.println("Unable to find decision service");
				return;
			}

			String initialReleaseName = IlrDecisionServiceHelper.getInitialReleaseName(service);
			IlrRelease depInitialRelease = (IlrRelease) ((IlrSessionEx) session).getBaselineNamed(project, initialReleaseName);

			if (depInitialRelease == null) {
				System.err.println("Initial release " + initialReleaseName+ " will be created for " + project.getName()+ " in Decision Service " + service.getName());

			}
			if (depInitialRelease == null || (depInitialRelease != null && !release.getParent().getName().equals(initialReleaseName))) {
				System.err.println("Set Initial release " + initialReleaseName+ " in release " + release.getName() + " for "+ project.getName() + " in Decision Service "+ service.getName());
			}
		} catch (IlrApplicationException e) {
			System.err.println("Unable to find decision service");
			e.printStackTrace();
		}

	}
	
	private void fixDependencies(IlrRuleProject project, IlrRelease release) {
		try {
			IlrRuleProject service = findDecisionService(project, release);
			if (service == null) {
				System.err.println("Unable to find decision service");
				return;
			}

			String initialReleaseName = IlrDecisionServiceHelper.getInitialReleaseName(service);
			IlrRelease depInitialRelease = (IlrRelease) ((IlrSessionEx) session).getBaselineNamed(project, initialReleaseName);

			if (depInitialRelease == null) {
				System.err.println("Initial release " + initialReleaseName+ " will be created for " + project.getName()+ " in Decision Service " + service.getName());

			}
			if (depInitialRelease == null || (depInitialRelease != null && !release.getParent().getName().equals(initialReleaseName))) {
				System.err.println("Set Initial release " + initialReleaseName+ " in release " + release.getName() + " for "+ project.getName() + " in Decision Service "+ service.getName());
			}
		} catch (IlrApplicationException e) {
			System.err.println("Unable to find decision service");
			e.printStackTrace();
		}

	}

	private Integer getId(IlrModelElement elt) {
		return ((IlrModelElementImpl) elt).getId();
	}

	/*
	 * create initial release TODO probably need to merge with a branch
	 */
	private IlrRelease createInitialRelease(IlrRuleProject dsProject,IlrRuleProject depProject, IlrRelease dsBranch) {
		try {
			IlrBrmPackage brm = session.getBrmPackage();
			ReleaseController releaseController = ((IlrSessionEx) session).getReleaseController();
			IlrRelease initialRelease = null;
			IlrElementHandle releaseHandle = session.createElement(brm.getRelease());
			IlrBranch currentBranch = IlrSessionHelper.getCurrentBaseline(session, depProject);
			initialRelease = (IlrRelease) session.getElementDetailsForThisHandle(releaseHandle);
			initialRelease.setRawValue(brm.getModelElement_Name(),IlrDecisionServiceHelper.INITIAL_RELEASE_NAME);

			// The initial release is a child of the current/main branch
			initialRelease.setRawValue(brm.getBaseline_Parent(), currentBranch);
			initialRelease.setRawValue(brm.getManagedBranch_Status(),releaseController.getInitialReleaseState());
			initialRelease.setRawValue(brm.getManagedBranch_Owner(),dsBranch.getOwner());
			initialRelease.setRawValue(brm.getManagedBranch_TargetDate(),dsBranch.getTargetDate());
			initialRelease.setRawValue(brm.getManagedBranch_Approvers(), null);
			initialRelease = (IlrRelease) ((IlrSessionEx) session).createDSBranch(dsProject, initialRelease, false);
			IlrProjectInfo projectInfo = initialRelease.getProjectInfo();
			
			if (projectInfo.getMigrationFlag() == 0) { // We don't want to set
														// it to 1 if its value
														// is already 3 (XOM
														// management)
				projectInfo.setRawValue(brm.getProjectInfo_MigrationFlag(), 1);
				IlrCommitableObject cObj = new IlrCommitableObject(projectInfo);
				cObj.setRootDetails(projectInfo);
				session.commit(null, cObj);
			}
			return initialRelease;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	private IlrRuleProject findDecisionService(IlrRuleProject prj,IlrRelease release) throws IlrApplicationException {
		IlrBrmPackage brm = session.getBrmPackage();
		IlrModelInfo mi = session.getModelInfo();

		EClass branchClass = (EClass) mi.getElementFromFQN(IlrModelInfo.getFQN(brm.getBranch()));
		// Find dependencies to this branch.
		DataFinder finder = DataFinder.createDataFinder(brm.getDependency());
		finder.addCriteria(brm.getDependency_ProjectName(), prj.getName());
		finder.addCriteria(brm.getDependency_BaselineName(), release.getName());
		List<IlrDependency> dependencies = finder.findDetails(session);
		for (IlrDependency dependency : dependencies) {
			IlrProjectInfo projectInfo = (IlrProjectInfo) session.getElementDetailsForThisHandle(dependency.getContainer());
			EClass prjInfoBslnClass = (EClass) mi.getElementFromFQN(IlrModelInfo.getFQN(projectInfo.getBaseline().eClass()));
			boolean isBranch = branchClass.isSuperTypeOf(prjInfoBslnClass);
			prj = projectInfo.getProject();
			if (prj.isDecisionService() == true) {
				return prj;
			}
		}
		return null;
	}

	public void checkProjectDependenciesInReleases(IlrRuleProject dsProject,String branch,Boolean fix) {
		try {
			List<IlrRelease> releaseHelper = IlrDecisionServiceHelper.getReleases(this.session, IlrDecisionServiceHelper.getInitialRelease(this.session, dsProject));
			for (IlrRelease release : releaseHelper) {
				List<String> releaseprojects2string = getDependenciesFromBaseline(release,"",fix);
				List<IlrActivity> activityHelper = IlrDecisionServiceHelper.getActivities(this.session, release);

				for (IlrActivity activity : activityHelper) {
					List<String> activityprojects2string = getDependenciesFromBaseline(activity,"",fix);

					compareDependencies(dsProject, release,releaseprojects2string, activity,activityprojects2string);
				}
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void compareDependencies(IlrRuleProject dsProject,
			IlrBaseline source, List<String> sourceProjectDependencies,
			IlrBaseline baseline, List<String> baselineProjectDependencies) {
		if (baselineProjectDependencies.size() != sourceProjectDependencies.size()) {
			ArrayList<String> workingDependencies = new ArrayList<String>();
			workingDependencies.addAll(sourceProjectDependencies);
			System.err.println("*********POSSIBLE ISSUE******** The branch '"+ source.getName()+ "' and the branch '"+ baseline.getName()+ "' don't have the same list of dependencies");
			if (baselineProjectDependencies.size() < workingDependencies.size()) {
				workingDependencies.removeAll(baselineProjectDependencies);
				System.err.println("The branch '" + baseline.getName()+ "' is missing the dependent project "+ workingDependencies);

			} else {
				baselineProjectDependencies.removeAll(workingDependencies);
				System.err.println("The branch '" + baseline.getName()+ "' has too many dependent project "+ baselineProjectDependencies);
			}
		}
	}
	
	public int occurences(String str, char temp){
		int charCount = 0;

		for( int i = 0; i < str.length( ); i++ )
		{

		    if( temp == str.charAt( i ))
		        charCount++;
		}
		
		return charCount;
		
		
	}

	private List<String> getDependenciesFromBaseline(IlrBaseline baseline, String path, Boolean fix)throws IlrApplicationException {
		//List<IlrBaseline> releaseprojects2 = IlrSessionHelper.computeDependentBaselines(this.session, baseline); This computes all dependencies of dependencies too
		IlrProjectInfo info = baseline.getProjectInfo();
		List<IlrDependency> releaseprojects = info.getDependencies();
		List<IlrBaseline> releaseprojects2 = new ArrayList<IlrBaseline>();
		List<String> releaseprojects2string = new ArrayList<String>();
		String o="  ";
		String off = new String(new char[this.occurences(path,'/')]).replace("\0", o);
		String destBranchName= baseline.getName();
		
		for (IlrDependency project2 : releaseprojects) {
			IlrRuleProject projectDep = IlrSessionHelper.getProjectNamed(session,project2.getProjectName());
			IlrBaseline depB=IlrSessionHelper.getBaselineNamed(session, projectDep, project2.getBaselineName());
			if(fix&&!depB.getName().equalsIgnoreCase(destBranchName)){
				System.out.println("Fixing dependency for "+projectDep.getName());
				      String tempBranchName= projectDep.getName()+ "_"+ destBranchName;
				      if(IlrSessionHelper.getBaselineNamed(session, projectDep, destBranchName) ==null){ 
				       // create a temporary backup baseline to avoid conflicts
				    	System.out.println("Dependant project "+projectDep.getName()+" not found in Project:"+baseline.getProject().getName()+" Branch:"+baseline.getName());
				    	System.out.println("Creating dependant project "+projectDep.getName());
				        session.setWorkingBaseline(depB);
				        IlrBranch temporaryBranch = session.createBranchFromCurrentState( tempBranchName, true);
				        renameBranch(temporaryBranch, destBranchName, new ArrayList<String>());
				     }
				    System.out.println("Setting Dependency to "+projectDep.getName()+":"+destBranchName);
					//change dependencies

				    IlrCommitableObject co = new IlrCommitableObject(info);
				    project2.setBaselineName(destBranchName);
				    co.addModifiedElement(session.getBrmPackage().getProjectInfo_Dependencies(), project2);
				    session.commit(co);
				
				
			}else
			releaseprojects2.add(depB);
		}	

		for (IlrBaseline rp : releaseprojects2) {
			String projectName=rp.getProject().getName();
			releaseprojects2string.add(projectName);
			if (projectPaths.containsKey(projectName)){
				List<String> newpath;
				if (projectPaths.get(projectName).containsKey(rp.getName())){
					newpath = projectPaths.get(projectName).get(rp.getName());
				}else{
					newpath =new ArrayList<String>();
				}
				newpath.add(path);
				projectPaths.get(projectName).put(rp.getName(),newpath);
				
				
			}else{
				List<String> newpath =new ArrayList<String>();
				newpath.add(path);
				Map<String, List<String> > temp = new HashMap<>();
				temp.put(rp.getName(),newpath);
				projectPaths.put(projectName, temp);
			}
		}
		
		System.out.println(off+" - Project: " + baseline.getProject().getName()+"* Branch: " + baseline.getName() + "->"+ releaseprojects2string);
		
		for (IlrBaseline rp : releaseprojects2) {
			getDependenciesFromBaseline(rp,path+'/'+rp.getProject().getName(),fix);
		}
		
		return releaseprojects2string;
	}
	
	public static void main(String[] args) {
		if (args.length < 4) System.err.println("Usage: CheckDS <serverURL> <datasource> <user> <password> [<projectName>]"+ " Without projectName will check the repository"+ "\n 1 check decision service dependencies"+ "\n 2 check decision service releases");
		
		String serverUrl = args[0];
		String datasource = args[1];
		String login = args[2];
		String password = args[3];
		String project = null;
		String branch = null;
		boolean fix = false;
		if (args.length > 4) if (args[4].equalsIgnoreCase("true"))fix=true;
		if (args.length > 5) project = args[5];
		if (args.length > 6) branch = args[6];
		
		
		IlrSessionFactory factory = new IlrRemoteSessionFactory();
		try {
			factory.connect(login, password, serverUrl, datasource);
			IlrSession session = factory.getSession();
			CheckCircularDependencies rtsClient = new CheckCircularDependencies(session);
			rtsClient.inspectRepository(project,branch,fix);
		
			
		} catch (IlrConnectException e) {
			e.printStackTrace();
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
		}
	}
}
