package com.ibm.rules.support;

import ilog.rules.teamserver.brm.IlrActivity;
import ilog.rules.teamserver.brm.IlrBaseline;
import ilog.rules.teamserver.brm.IlrBranch;
import ilog.rules.teamserver.brm.IlrBrmPackage;
import ilog.rules.teamserver.brm.IlrChangeAct;
import ilog.rules.teamserver.brm.IlrDependency;
import ilog.rules.teamserver.brm.IlrModelElement;
import ilog.rules.teamserver.brm.IlrProjectInfo;
import ilog.rules.teamserver.brm.IlrRelease;
import ilog.rules.teamserver.brm.IlrRuleProject;
import ilog.rules.teamserver.brm.impl.IlrDependencyImpl;
import ilog.rules.teamserver.brm.impl.IlrModelElementImpl;
import ilog.rules.teamserver.client.IlrRemoteSessionFactory;
import ilog.rules.teamserver.model.IlrApplicationException;
import ilog.rules.teamserver.model.IlrCommitableObject;
import ilog.rules.teamserver.model.IlrConnectException;
import ilog.rules.teamserver.model.IlrCyclicBaselineDependencyException;
import ilog.rules.teamserver.model.IlrElementHandle;
import ilog.rules.teamserver.model.IlrModelInfo;
import ilog.rules.teamserver.model.IlrObjectNotFoundException;
import ilog.rules.teamserver.model.IlrSession;
import ilog.rules.teamserver.model.IlrSessionEx;
import ilog.rules.teamserver.model.IlrSessionFactory;
import ilog.rules.teamserver.model.IlrSessionHelper;
import ilog.rules.teamserver.model.IlrSessionHelperEx;
import ilog.rules.teamserver.model.finders.DataFinder;
import ilog.rules.teamserver.model.governance.ReleaseController;
import ilog.rules.teamserver.model.impl.IlrDecisionServiceHelper;
import ilog.rules.teamserver.model.permissions.IlrPermissionException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;



public class DcCheckDS {
	boolean dgfOverride = true;
	boolean debug = true;
	protected IlrSession session;
	private boolean applyFix;
	private static String version = "V2.1";
	static Map<String, Map<String, List<String>> > projectPaths = new HashMap<>();

	public DcCheckDS(IlrSession session, boolean applyFix) {
		this.session = session;
		this.applyFix = applyFix;
		System.out.println("Check Decision Services " + version+ ": fix mode is " + (applyFix ? "on" : "off"));
	}

	@SuppressWarnings("unchecked")
	public List<String> getParentProjects(IlrSession session,
			IlrProjectInfo projectinfo, String depProject) {
		List<String> result = new ArrayList<String>();
		if (depProject.isEmpty())
			return result;
		try {
			List<IlrDependencyImpl> releaseprojects = projectinfo
					.getDependencies();

			for (IlrDependencyImpl project : releaseprojects) {
				if (depProject.equals(project.getProjectName())) {
					result.add(projectinfo.getProject().getName());
				}
			}

			for (IlrDependencyImpl project2 : releaseprojects) {

				String baselinename = project2.getBaselineName();
				String projectname = project2.getProjectName();
				IlrRuleProject projectDep = IlrSessionHelper.getProjectNamed(
						session, projectname);
				IlrBaseline baseline = IlrSessionHelper.getBaselineNamed(
						session, projectDep, baselinename);
				IlrProjectInfo info = baseline.getProjectInfo();

				result.addAll(getParentProjects(session, info, depProject));
			}
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
		}
		result = new ArrayList<String>(new HashSet<String>(result));
		return result;
	}

	public static void main(String[] args) {
		if (args.length < 4) {
			System.err
					.println("Usage: CheckDS <serverURL> <datasource> <user> <password> [<projectName>]"
							+ " Without projectName will check the repository"
							+ "\n 1 check decision service dependencies"
							+ "\n 2 check decision service releases");
		}
		String serverUrl = args[0];
		String datasource = args[1];
		String login = args[2];
		String password = args[3];
		String project = null;

		if (args.length > 4) {
			project = args[4];
		}

		boolean doFix = false;
		String dsName = null;

		if (project != null)
			if (project.endsWith("-fix")) {
				doFix = true;
				dsName = project.substring(0, project.length() - 4);
			} else {
				dsName = project;
			}
		IlrSessionFactory factory = new IlrRemoteSessionFactory();
		try {
			factory.connect(login, password, serverUrl, datasource);
			IlrSession session = factory.getSession();
			DcCheckDS rtsClient = new DcCheckDS(session, doFix);
			rtsClient.inspectRepository(dsName);/////////////////////////////////////
			
		} catch (IlrConnectException e) {
			e.printStackTrace();
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
		}
	}

	public void inspectRepository(String dsName)throws IlrObjectNotFoundException {
		if ((dsName != null) && (!dsName.isEmpty())) {
			IlrRuleProject dsProject = IlrSessionHelper.getProjectNamed(this.session, dsName);
			checkProject(dsProject);
		} else {
			List<IlrRuleProject> allProjects = IlrSessionHelper.getProjects(this.session);
			for (IlrRuleProject ilrRuleProject : allProjects) {
				System.out.println("Checking "+ilrRuleProject.getName());
				checkProject(ilrRuleProject);
			}
		}
	}

	public void checkProject(IlrRuleProject ruleproject) {
		try {
			
			if (ruleproject.isDecisionService()) {
				if (IlrDecisionServiceHelper.isDGFProject(session, ruleproject)&&!dgfOverride) {
					checkProjectDependenciesInReleases(ruleproject);
				} else {
					checkProjectDependenciesInBranch(ruleproject);
				}
			}
			checkReleaseParent(ruleproject);

		} catch (Exception e) {
			System.out.println("- Project: " + (ruleproject != null ? (ruleproject.getName() + "Unable to process this project"): "Not found"));
			e.printStackTrace();
		}
	}
	
	public void printProjectPaths(String n,boolean debug){
	
	for (String p : projectPaths.keySet()) {
		boolean err=false;
		String o="  ";
		if(!debug && projectPaths.get(p).keySet().size()>1){err=true;System.out.println("Found Conflicting Dependencies in "+n);}
		if(debug||err)System.out.println((new String(new char[1]).replace("\0", o))+"Dependency to "+p);
		
		for (String b : projectPaths.get(p).keySet()){
			//if(debug||err)System.out.println((new String(new char[1]).replace("\0", o))+"Branch "+b);
			for (String path : projectPaths.get(p).get(b))
				if(debug||err)System.out.println((new String(new char[2]).replace("\0", o))+"Project \""+path+ "\" Branch="+b);
			}
		
		} 
	}


	private void checkProjectDependenciesInBranch(IlrRuleProject ruleproject) {
		try {
			
			projectPaths = new HashMap<>();
			
			IlrBaseline mainBranch = IlrSessionHelper.getCurrentBaseline(session, ruleproject) ;
				List<String> mainDependentProjects = getDependenciesFromBaseline(mainBranch,ruleproject.getName());
				List<IlrBaseline> branches = IlrSessionHelper.getBaselines(session, ruleproject);
				
				printProjectPaths(ruleproject.getName()+':'+mainBranch.getName(),debug);

				for (IlrBaseline baseline: branches) {
					projectPaths = new HashMap<>();
					if (!baseline.equals( mainBranch) && !IlrDecisionServiceHelper.isDGFBranch(baseline) && baseline.isBranch() )
					{
						List<String> baselineDependentProjects = getDependenciesFromBaseline(baseline,ruleproject.getName());
						
						printProjectPaths(ruleproject.getName()+':'+baseline.getName(),false);
						
						//compareDependencies(ruleproject, mainBranch,mainDependentProjects, baseline, baselineDependentProjects);
					}
				}

		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}

	public void checkReleaseParent(IlrRuleProject dsProject) {
		try {
			List<IlrBaseline> baselines = IlrSessionHelper.getBaselines(session, dsProject);
			for (IlrBaseline baseline : baselines) {
				// searching for release
				if (baseline.isDecisionServiceBranch()&& session.getBrmPackage().getRelease().isSuperTypeOf(baseline.eClass())) {
					IlrRelease release = (IlrRelease) baseline;
					if (release.isInitialRelease()&& !release.getName().contains("odm_initialRelease_")) {
						if (release.getParent() != null&& release.getParent().getParent() == null) {
							System.err.println("* release "+ release.getName()+ " does not have Initial Release for parent");
							fixReleaseParent(dsProject, release);
						}
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void fixReleaseParent(IlrRuleProject project, IlrRelease release) {
		try {
			IlrRuleProject service = findDecisionService(project, release);
			if (service == null) {
				System.err.println("Unable to find decision service");
				return;
			}

			String initialReleaseName = IlrDecisionServiceHelper
					.getInitialReleaseName(service);
			IlrRelease depInitialRelease = (IlrRelease) ((IlrSessionEx) session)
					.getBaselineNamed(project, initialReleaseName);

			if (depInitialRelease == null) {
				System.err.println("Initial release " + initialReleaseName
						+ " will be created for " + project.getName()
						+ " in Decision Service " + service.getName());
				if (this.applyFix) {
					depInitialRelease = createInitialRelease(service, project,
							release);
				}
			}
			if (depInitialRelease == null
					|| (depInitialRelease != null && !release.getParent()
							.getName().equals(initialReleaseName))) {
				System.err.println("Set Initial release " + initialReleaseName
						+ " in release " + release.getName() + " for "
						+ project.getName() + " in Decision Service "
						+ service.getName());
				if (this.applyFix && depInitialRelease != null) { /*
																 * currently
																 * print the SQL
																 * query to fix
																 * the
																 * dependency
																 * issue. TODO
																 * tried to
																 * commit a
																 * simple change
																 * of parent but
																 * a release is
																 * not a
																 * commitable
																 * object tried
																 * to delete the
																 * release and
																 * recreate it
																 * but would
																 * probably lose
																 * changes doing
																 * that. how to
																 * update a
																 * release
																 * parent using
																 * API?
																 */

					System.err.println("run UPDATE BASELINE SET PARENT = "
							+ getId(depInitialRelease) + " WHERE ID = "
							+ getId(release) + " AND CONTAINER = "
							+ getId(project));
				}
			}
		} catch (IlrApplicationException e) {
			System.err.println("Unable to find decision service");
			e.printStackTrace();
		}

	}

	private Integer getId(IlrModelElement elt) {
		return ((IlrModelElementImpl) elt).getId();
	}

	/*
	 * create initial release TODO probably need to merge with a branch
	 */
	private IlrRelease createInitialRelease(IlrRuleProject dsProject,
			IlrRuleProject depProject, IlrRelease dsBranch) {
		try {
			IlrBrmPackage brm = session.getBrmPackage();
			ReleaseController releaseController = ((IlrSessionEx) session)
					.getReleaseController();
			IlrRelease initialRelease = null;
			IlrElementHandle releaseHandle = session.createElement(brm
					.getRelease());
			IlrBranch currentBranch = IlrSessionHelper.getCurrentBaseline(
					session, depProject);
			initialRelease = (IlrRelease) session
					.getElementDetailsForThisHandle(releaseHandle);
			initialRelease.setRawValue(brm.getModelElement_Name(),
					IlrDecisionServiceHelper.INITIAL_RELEASE_NAME);

			// The initial release is a child of the current/main branch
			initialRelease.setRawValue(brm.getBaseline_Parent(), currentBranch);
			initialRelease.setRawValue(brm.getManagedBranch_Status(),
					releaseController.getInitialReleaseState());
			initialRelease.setRawValue(brm.getManagedBranch_Owner(),
					dsBranch.getOwner());
			initialRelease.setRawValue(brm.getManagedBranch_TargetDate(),
					dsBranch.getTargetDate());
			initialRelease.setRawValue(brm.getManagedBranch_Approvers(), null);
			initialRelease = (IlrRelease) ((IlrSessionEx) session)
					.createDSBranch(dsProject, initialRelease, false);
			IlrProjectInfo projectInfo = initialRelease.getProjectInfo();
			if (projectInfo.getMigrationFlag() == 0) { // We don't want to set
														// it to 1 if its value
														// is already 3 (XOM
														// management)
				projectInfo.setRawValue(brm.getProjectInfo_MigrationFlag(), 1);
				IlrCommitableObject cObj = new IlrCommitableObject(projectInfo);
				cObj.setRootDetails(projectInfo);
				session.commit(null, cObj);
			}
			return initialRelease;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	private IlrRuleProject findDecisionService(IlrRuleProject prj,
			IlrRelease release) throws IlrApplicationException {
		IlrBrmPackage brm = session.getBrmPackage();
		IlrModelInfo mi = session.getModelInfo();

		EClass branchClass = (EClass) mi.getElementFromFQN(IlrModelInfo
				.getFQN(brm.getBranch()));
		// Find dependencies to this branch.
		DataFinder finder = DataFinder.createDataFinder(brm.getDependency());
		finder.addCriteria(brm.getDependency_ProjectName(), prj.getName());
		finder.addCriteria(brm.getDependency_BaselineName(), release.getName());
		List<IlrDependency> dependencies = finder.findDetails(session);
		for (IlrDependency dependency : dependencies) {
			IlrProjectInfo projectInfo = (IlrProjectInfo) session
					.getElementDetailsForThisHandle(dependency.getContainer());
			EClass prjInfoBslnClass = (EClass) mi
					.getElementFromFQN(IlrModelInfo.getFQN(projectInfo
							.getBaseline().eClass()));
			boolean isBranch = branchClass.isSuperTypeOf(prjInfoBslnClass);
			prj = projectInfo.getProject();
			if (prj.isDecisionService() == true) {
				return prj;
			}
		}
		return null;
	}

	public void checkProjectDependenciesInReleases(IlrRuleProject dsProject) {
		try {
			List<IlrRelease> releaseHelper = IlrDecisionServiceHelper
					.getReleases(this.session, IlrDecisionServiceHelper
							.getInitialRelease(this.session, dsProject));
			for (IlrRelease release : releaseHelper) {
				List<String> releaseprojects2string = getDependenciesFromBaseline(release,"");

				List<IlrActivity> activityHelper = IlrDecisionServiceHelper
						.getActivities(this.session, release);

				for (IlrActivity activity : activityHelper) {

					List<String> activityprojects2string = getDependenciesFromBaseline(activity,"");

					compareDependencies(dsProject, release,
							releaseprojects2string, activity,
							activityprojects2string);

				}
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private void compareDependencies(IlrRuleProject dsProject,
			IlrBaseline source, List<String> sourceProjectDependencies,
			IlrBaseline baseline, List<String> baselineProjectDependencies) {
		if (baselineProjectDependencies.size() != sourceProjectDependencies
				.size()) {
			ArrayList<String> workingDependencies = new ArrayList<String>();
			workingDependencies.addAll(sourceProjectDependencies);
			System.err
					.println("*********POSSIBLE ISSUE******** The branch '"
							+ source.getName()
							+ "' and the branch '"
							+ baseline.getName()
							+ "' don't have the same list of dependencies");
			if (baselineProjectDependencies.size() < workingDependencies.size()) {
				workingDependencies.removeAll(baselineProjectDependencies);
				System.err.println("The branch '" + baseline.getName()
						+ "' is missing the dependent project "
						+ workingDependencies);
				for (String depProjectName : workingDependencies) {
					fixProject(dsProject.getName(), source, baseline.getName(),
							depProjectName);
				}
			} else {
				baselineProjectDependencies.removeAll(workingDependencies);
				System.err.println("The branch '" + baseline.getName()
						+ "' has too many dependent project "
						+ baselineProjectDependencies);
			}

		}
	}
	
	public int occurences(String str, char temp){
		int charCount = 0;

		for( int i = 0; i < str.length( ); i++ )
		{
		    

		    if( temp == str.charAt( i ))
		        charCount++;
		}
		
		return charCount;
		
		
	}

	private List<String> getDependenciesFromBaseline(IlrBaseline baseline, String path)throws IlrObjectNotFoundException {
		List<IlrBaseline> releaseprojects2 = IlrSessionHelper.computeDependentBaselines(this.session, baseline);
		List<String> releaseprojects2string = new ArrayList<String>();
		String o="  ";
		String off = new String(new char[this.occurences(path,'/')]).replace("\0", o);

		for (IlrBaseline rp : releaseprojects2) {
			String projectName=rp.getProject().getName();
			releaseprojects2string.add(projectName);
			if (projectPaths.containsKey(projectName)){
				List<String> newpath;
				if (projectPaths.get(projectName).containsKey(rp.getName())){
					newpath = projectPaths.get(projectName).get(rp.getName());
				}else{
					newpath =new ArrayList<String>();
				}
				newpath.add(path);
				projectPaths.get(projectName).put(rp.getName(),newpath);
				
				
			}else{
				List<String> newpath =new ArrayList<String>();
				newpath.add(path);
				Map<String, List<String> > temp = new HashMap<>();
				temp.put(rp.getName(),newpath);
				projectPaths.put(projectName, temp);
			}
		}
		
		//System.out.println(off+" - Project: " + baseline.getProject().getName());
		//System.out.println(off+o+"* Branch: " + baseline.getName() + "->"+ releaseprojects2string);
		
		System.out.println(off+" - Project: " + baseline.getProject().getName()+"* Branch: " + baseline.getName() + "->"+ releaseprojects2string);
		
		for (IlrBaseline rp : releaseprojects2) {
			getDependenciesFromBaseline(rp,path+'/'+rp.getProject().getName());
		}
		
		
		return releaseprojects2string;
	}

	public void fixProject(String dsName, IlrBaseline release,
			String activityName, String projectToAdd) {
		IlrBrmPackage brm = this.session.getBrmPackage();

		try {
			IlrProjectInfo releaseInfo = release.getProjectInfo();

			List<String> targetProjects = getParentProjects(this.session,
					releaseInfo, projectToAdd);

			System.err
					.println(projectToAdd
							+ (this.applyFix ? " will " : " should ")
							+ " be added to "
							+ getParentProjects(this.session, releaseInfo,
									projectToAdd));

			if (!this.applyFix) {
				return;
			}
			IlrRuleProject dsProject = IlrSessionHelper.getProjectNamed(
					this.session, dsName);

			IlrRuleProject depProject = IlrSessionHelper.getProjectNamed(
					this.session, projectToAdd);

			IlrRelease depRelease = (IlrRelease) ((IlrSessionEx) this.session)
					.getBaselineNamed(depProject, release.getName());

			for (String targetProject : targetProjects) {
				IlrActivity activity = (IlrActivity) IlrSessionHelper
						.getBaselineNamed(this.session, IlrSessionHelper
								.getProjectNamed(this.session, targetProject),
								activityName);
				IlrSessionHelperEx.addBaselineDependency(this.session,
						activity, projectToAdd, activity.getName());
				IlrElementHandle activityHandle = this.session
						.createElement(brm.getChangeAct());
				IlrChangeAct newActivity = (IlrChangeAct) this.session
						.getElementDetailsForThisHandle(activityHandle);
				newActivity.setRawValue(brm.getModelElement_Name(),
						activity.getName());
				newActivity.setRawValue(brm.getBaseline_Parent(), depRelease);
				newActivity.setRawValue(brm.getManagedBranch_Status(),
						activity.getStatus());
				newActivity.setRawValue(brm.getManagedBranch_Owner(),
						activity.getOwner());
				newActivity.setRawValue(brm.getManagedBranch_TargetDate(),
						activity.getTargetDate());
				newActivity.setRawValue(brm.getManagedBranch_Approvers(), null);
				newActivity = (IlrChangeAct) ((IlrSessionEx) this.session)
						.createDSBranch(dsProject, newActivity);
			}
		} catch (IlrObjectNotFoundException e) {
			e.printStackTrace();
		} catch (IlrCyclicBaselineDependencyException e) {
			e.printStackTrace();
		} catch (IlrPermissionException e) {
			e.printStackTrace();
		} catch (IlrApplicationException e) {
			e.printStackTrace();
		}
	}
}
