package com.ibm.rules.support;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.ibm.rules.decisioncenter.BRMServerClientLegacy;

import ilog.rules.teamserver.brm.IlrBaseline;
import ilog.rules.teamserver.brm.IlrBrmPackage;
import ilog.rules.teamserver.brm.IlrRuleProject;
import ilog.rules.teamserver.model.IlrApplicationException;
import ilog.rules.teamserver.model.IlrObjectNotFoundException;
import ilog.rules.teamserver.model.IlrSessionHelper;
import ilog.rules.teamserver.model.permissions.IlrPermissionException;
import ilog.rules.teamserver.model.testing.IlrTestingHelper;
import ilog.rules.teamserver.validation.IlrTestSuite;
import ilog.rules.teamserver.validation.IlrValidationPackage;

public class TestSuiteRepair extends BRMServerClientLegacy {
	public String project;
	public IlrValidationPackage valPackage;
	public IlrBrmPackage brm;

	protected void init() throws Exception {
		this.valPackage = getSession().getValidationPackage();
		this.brm = getSession().getBrmPackage();
	}

	public void inspectRepository(String dsName, boolean fix) throws IlrObjectNotFoundException {
		try {
			init();
		} catch (Exception e) {
			e.printStackTrace();
			return;
		}
		if ((dsName != null) && (!dsName.isEmpty())) {
			IlrRuleProject dsProject = IlrSessionHelper.getProjectNamed(getSession(), dsName);
			checkProject(dsProject, fix);
		} else {
			List<IlrRuleProject> allProjects = IlrSessionHelper.getProjects(getSession());
			for (IlrRuleProject ilrRuleProject : allProjects) {
				checkProject(ilrRuleProject, fix);
			}
		}
	}

	private void checkProject(IlrRuleProject dsProject, boolean fix) {
		try {
			if (!dsProject.isDecisionService())
				return;

			List<IlrBaseline> baselines;
			baselines = IlrSessionHelper.getBaselines(getSession(), dsProject);
			System.out.println("Checking " + dsProject.getName() );
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			// when out of network loading the dtd will fail, so ignore the DTD document by default
			factory.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
			DocumentBuilder builder = factory.newDocumentBuilder();
			for (IlrBaseline ilrBaseline : baselines) {
				if ( !ilrBaseline.isBranch())
					continue;
				getSession().setWorkingBaseline(ilrBaseline);
				try {
					List<IlrTestSuite> suites = IlrTestingHelper.getTestSuites(getSession(), ilrBaseline, false);
					boolean found = false;
					for (IlrTestSuite suite : suites) {
						ByteArrayInputStream is = new ByteArrayInputStream( suite.getDescriptorXml());
						Document doc = builder.parse(is);
						NodeList entries = doc.getElementsByTagName("entry");
						for (int i = 0; i < entries.getLength(); i++) {
							Element e = (Element) entries.item(i);
							if (e.getAttribute("key").equals("testable.execution.details")) {
								if (!found ) System.out.println("\tBranch: "+ilrBaseline.getName() );
								System.out.println( "\t\t- Test suite: ["+ suite.getName() + "] uses {"+ e.getTextContent()+"}");
								found = true;
								break;
							}							
						}
					}
				} catch (IlrApplicationException | SAXException | IOException e) {
					e.printStackTrace();
				}
				
			}
		} catch (

		IlrObjectNotFoundException e) {
			e.printStackTrace();
		} catch (IlrPermissionException e1) {
			e1.printStackTrace();
		} catch (ParserConfigurationException e1) {
			e1.printStackTrace();
		}
	}

	public static void main(final String[] args) {
		try {
			(new TestSuiteRepair()).doExecute(args);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	protected void execute(String[] otherArgs) {
		String prjName = null;
		boolean fixMode = false;
		int i = 0;
		int valueIndex = -1; 
		for (String param : otherArgs) {
			if (valueIndex ==i)
			{
				i++;
				continue;
			}
			valueIndex = i+1;
			String value = otherArgs[valueIndex];
			switch (param) {
			case PRJ_ARGNAME:
				prjName = value;
				break;
			case FIX_MODE:
				fixMode = value.equalsIgnoreCase("true");
				break;
			default:
				System.out.println(" unexpected argument " + param);
				usage();
				return;
			}
			i++;
		}
		
		try {
				inspectRepository(prjName, fixMode);

			} catch (Exception e){
				e.printStackTrace();
			}
	}

	private final static String PRJ_ARGNAME = "-project";
	private final static String FIX_MODE = "-fix";

	@Override
	protected String usageArgs() {
		return PRJ_ARGNAME + " <dsPrjName> " + FIX_MODE + "true|false"
				+ "\tcheck the bom path dependencies in all project  or a given project display the problems and repair if fix mode is true";

	}
}
