package com.ibm.rules.support;

import ilog.rules.teamserver.brm.IlrBrmPackage;
import ilog.rules.teamserver.model.IlrElementDetails;
import ilog.rules.teamserver.model.IlrCommitableObject;
import ilog.rules.teamserver.validation.IlrTestReport;
import com.ibm.bdsl.util.Pair;
import java.util.Iterator;
import java.util.Collection;
import ilog.rules.teamserver.model.IlrElementHandle;
import ilog.rules.teamserver.model.IlrSessionHelper;
import ilog.rules.teamserver.brm.IlrDependency;
import java.util.ArrayList;
import ilog.rules.teamserver.brm.IlrProjectInfo;
import ilog.rules.teamserver.brm.IlrBaseline;
import java.util.List;
import ilog.rules.teamserver.model.IlrElementSummary;
import ilog.rules.teamserver.model.IlrSearchCriteria;
import ilog.rules.teamserver.model.IlrDefaultSearchCriteria;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.FileReader;
import ilog.rules.teamserver.model.IlrSessionFactory;
import ilog.rules.teamserver.client.IlrRemoteSessionFactory;
import ilog.rules.teamserver.validation.IlrValidationPackage;
import ilog.rules.teamserver.model.IlrSession;

public class testReportDependencies
{
    protected String serverUrl;
    protected String datasource;
    protected String login;
    protected String password;
    public String project;
    public IlrSession session;
    public IlrValidationPackage valPackage;
    
    public testReportDependencies(final String serverUrl, final String datasource, final String login, final String password) {
        this.serverUrl = serverUrl;
        this.datasource = datasource;
        this.login = login;
        this.password = password;
    }
    
    protected void init() throws Exception {
        final IlrSessionFactory factory = (IlrSessionFactory)new IlrRemoteSessionFactory();
        factory.connect(this.login, this.password, this.serverUrl, this.datasource);
        (this.session = factory.getSession()).beginUsage();
        this.valPackage = this.session.getValidationPackage();
    }
    
    public List<IlrProjectInfo> getSubProjectInfoWithDependency(final IlrBaseline f, final String dependency) throws Exception {
        final List<IlrProjectInfo> ret = new ArrayList<IlrProjectInfo>();
        final List<IlrDependency> releaseprojects2 = (List<IlrDependency>)f.getProjectInfo().getDependencies();
        for (final IlrDependency rp : releaseprojects2) {
            if (rp.getProjectName().equals(dependency)) {
                ret.add(f.getProjectInfo());
            }
            ret.addAll(this.getSubProjectInfoWithDependency(IlrSessionHelper.getBaselineNamed(this.session, (IlrElementHandle)IlrSessionHelper.getProjectNamed(this.session, rp.getProjectName()), rp.getBaselineName()), dependency));
        }
        return ret;
    }
    
    public List<IlrDependency> getDependencies(final IlrBaseline f, final String dependency) throws Exception {
        final List<IlrDependency> ret = new ArrayList<IlrDependency>();
        final List<IlrDependency> releaseprojects2 = (List<IlrDependency>)f.getProjectInfo().getDependencies();
        for (final IlrDependency rp : releaseprojects2) {
            if (rp.getProjectName().equals(dependency)) {
                ret.add(rp);
            }
            ret.addAll(this.getDependencies(IlrSessionHelper.getBaselineNamed(this.session, (IlrElementHandle)IlrSessionHelper.getProjectNamed(this.session, rp.getProjectName()), rp.getBaselineName()), dependency));
        }
        return ret;
    }
    
    public Pair<List<IlrProjectInfo>, List<IlrDependency>> getBoth(final IlrBaseline f, final String dependency) throws Exception {
        final List<IlrProjectInfo> ret1 = new ArrayList<IlrProjectInfo>();
        final List<IlrDependency> ret2 = new ArrayList<IlrDependency>();
        final List<IlrDependency> releaseprojects2 = (List<IlrDependency>)f.getProjectInfo().getDependencies();
        for (final IlrDependency rp : releaseprojects2) {
            if (rp.getProjectName().equals(dependency)) {
                ret1.add(f.getProjectInfo());
                ret2.add(rp);
            }
            final Pair<List<IlrProjectInfo>, List<IlrDependency>> p = this.getBoth(IlrSessionHelper.getBaselineNamed(this.session, (IlrElementHandle)IlrSessionHelper.getProjectNamed(this.session, rp.getProjectName()), rp.getBaselineName()), dependency);
            ret1.addAll((Collection<? extends IlrProjectInfo>)p.first);
            ret2.addAll((Collection<? extends IlrDependency>)p.second);
        }
        return (Pair<List<IlrProjectInfo>, List<IlrDependency>>)new Pair((Object)ret1, (Object)ret2);
    }
    
    //find/print all test reports
    public void query(final String project, final boolean fix) throws Exception {
        this.init();
        final List<IlrBaseline> fixableBaselines = new ArrayList<IlrBaseline>();
        System.out.println("Finding Test Reports:");
        final IlrDefaultSearchCriteria criteria = new IlrDefaultSearchCriteria(this.valPackage.getTestReport());
        final List<?> summaries = (List<?>)this.session.findElements((IlrSearchCriteria)criteria);
        System.out.println(String.valueOf(summaries.size()) + " reports found.");
        if (project != null) {
            System.out.println("Searching for reports with dependency to " + project + ":");
        }
        else {
            System.out.println("Printing all test reports");
        }
        if (summaries.size() == 0) {
            return;
        }
        for (int i = 0; i < summaries.size(); ++i) {
            final IlrTestReport report = (IlrTestReport)summaries.get(i);
            final IlrBaseline origbaseline = report.getOriginalBaseline();
            final IlrBaseline snapshotbaseline = (IlrBaseline)report.getReportBaseline();
            if (!origbaseline.getProject().getName().equals(project)) {
                final List<String> snapshotdependencies2string = new ArrayList<String>();
                final List<IlrBaseline> snapshotdependencies = (List<IlrBaseline>)IlrSessionHelper.computeDependentBaselines(this.session, snapshotbaseline);
                boolean b = false;
                for (final IlrBaseline d : snapshotdependencies) {
                    if (project != null && project.equals(d.getProject().getName())) {
                        b = true;
                    }
                    snapshotdependencies2string.add(d.getProject().getName());
                }
                if (b || project == null) {
                    if (fix) {
                        fixableBaselines.add(snapshotbaseline);
                    }
                    System.out.println("\t" + report.getName() + " ProjectName: " + origbaseline.getProject().getName() + "(" + origbaseline.getName() + ")" + "(" + snapshotbaseline.getName() + ")" + "(" + snapshotbaseline.toIdString() + ")");
                    System.out.println("\t\tDependencies: " + snapshotdependencies2string);
                }
            }
        }
        if (fix && fixableBaselines.size() > 0) {
            System.out.println("Removing dependency from " + fixableBaselines.size() + " matching test reports:");
            for (final IlrBaseline f : fixableBaselines) {
                final Pair<List<IlrProjectInfo>, List<IlrDependency>> both = this.getBoth(f, project);
                for (int j = 0; j < ((List)both.first).size(); ++j) {
                    final IlrDependency dep = (IlrDependency) ((List)both.second).get(j);
                    final IlrCommitableObject cProjectInfo = new IlrCommitableObject((IlrElementHandle)((List)both.first).get(j));
                    cProjectInfo.setRootDetails((IlrElementDetails)((List)both.first).get(j));
                    final IlrSession session2 = ((IlrElementHandle) ((List)both.first).get(j)).getSession();
                    final IlrBrmPackage brmPackage = session2.getBrmPackage();
                    cProjectInfo.addDeletedElement(brmPackage.getProjectInfo_Dependencies(), (IlrElementDetails)dep);
                    session2.commit(cProjectInfo);
                }
                System.out.println("Successfully removed dependency from " + f.getName());
            }
        }
    }
    
    public static void main(final String[] args) {
        try {
            if (args.length < 4 || args.length > 6) {
                System.err.println("Improper arguments");
                System.exit(0);
            }
            final String serverUrl = args[0];
            final String datasource = args[1];
            final String login = args[2];
            final String password = args[3];
            String project = null;
            boolean fix = false;
            if (args.length > 4) {
                project = args[4];
            }
            if (args.length > 5 && args[5].equalsIgnoreCase("true")) {
                fix = true;
            }
            final testReportDependencies rtsClient = new testReportDependencies(serverUrl, datasource, login, password);
            rtsClient.query(project, fix);
            System.out.println("Complete");
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
}